if (ZOMGBuffTehRaid) then
	z:Print("Installation error, duplicate copy of ZOMGBuffs_BuffTehRaid (Addons\ZOMGBuffs\ZOMGBuffs_BuffTehRaid and Addons\ZOMGBuffs_BuffTehRaid)")
	return
end

local L = AceLibrary("AceLocale-2.2"):new("ZOMGBuffTehRaid")
local R = AceLibrary("ZOMGReagents-1.0")
local BS = AceLibrary("Babble-Spell-2.2")
local roster = AceLibrary("Roster-2.1")
local playerClass
local template
local mainSingle, mainGroup

local wow24 = GetBuildInfo() < "1.0.0" or GetBuildInfo() >= "2.4.0"

local z = ZOMGBuffs
local zg = z:NewModule("ZOMGBuffTehRaid")
ZOMGBuffTehRaid = zg

z:CheckVersion("$Revision: 64927 $")

local new, del, deepDel, copy = z.new, z.del, z.deepDel, z.copy

local function getOption(k)
	return zg.db.char[k]
end

local function setOption(k, v, update)
	if (not v or v == 0) then
		zg.db.char[k] = nil
	else
		zg.db.char[k] = v
	end
	if (update) then
		z:CheckForChange(zg)
	end
end

local function getPrelude(k)
	return zg.db.char.rebuff[k] or 0
end

local function setPrelude(k, v)
	if (not v or v == 0) then
		zg.db.char.rebuff[k] = nil
	else
		zg.db.char.rebuff[k] = v
	end
	z:CheckForChange(zg)
end

zg.consoleCmd = "ZOMGBuffTehRaid"
zg.options = {
	type = 'group',
	order = 2,
	name = "|cFFFF8080Z|cFFFFFF80O|cFF80FF80M|cFF8080FFG|cFFFFFFFFBuffTehRaid|r",
	desc = L["Group Buff Configuration"],
	handler = zg,
	args = {
		template = {
			type = 'group',
			name = L["Templates"],
			desc = L["Template configuration"],
			order = 1,
			hidden = function() return not zg:IsModuleActive() end,
			args = {
			}
		},
		groups = {
			type = 'group',
			name = L["Groups"],
			desc = L["Select the groups to buff"],
			order = 2,
			hidden = function() return not zg:IsModuleActive() end,
			args = {
				auto = {
					type = 'toggle',
					name = L["Auto-Assign"],
					desc = L["Auto assign sensible group assignment based the order of your name alphabilically compared to others of your class. All going well, and all using ZOMGBuffs and everyone should end up with different assignments without need for discussion"],
					get = getOption,
					set = function(k,v) setOption(k,v) zg:AutoGroupAssignment() end,
					passValue = "autogroups",
					order = 1,
				},
				space = {
					type = 'header',
					name = " ",
					order = 2,
				},
			},
		},
		behaviour = {
			type = 'group',
			name = L["Behaviour"],
			desc = L["Group buffing behaviour"],
			order = 201,
			hidden = function() return not zg:IsModuleActive() end,
			args = {
				groupcast = {
					type = 'range',
					name = L["Minimum Group"],
					desc = L["How many players of a group must need a buff before the group version is cast"],
					get = getOption,
					set = function(k,v) setOption(k,v,true) end,
					passValue = "groupcast",
					order = 1,
					min = 1,
					max = 5,
					step = 1,
				},
				rebuff = {
					type = 'range',
					name = L["Expiry Prelude"],
					desc = L["Default rebuff prelude for all group buffs"],
					func = timeFunc,
					order = 2,
					get = getPrelude,
					set = function(k,v) setOption(k,v,true) end,
					passValue = "default",
					min = 0,
					max = 15 * 60,
					step = 5,
					bigStep = 60,
					order = 2,
				},
			},
		},
	}
}

do
	local disfunc = function() return zg.db.char.autogroups end
	local getG = function(i) return zg.db.char.groups[i] end
	local setG = function(i,n) zg.db.char.groups[i] = n and true or nil z:CheckForChange(zg) end
	local args = zg.options.args.groups.args
	for i = 1,8 do
		args[i] = {
			type = 'toggle',
			name = format(L["Group %d"], i),
			desc = L["Enable buffing of this group"],
			order = i + 10,
			get = getG,
			set = setG,
			disabled = disfunc,
			passValue = i,
		}
	end
end

local function getOption(v)
	return zg.db.char[v]
end

local function setOption(v, n)
	zg.db.char[v] = n
end

-- ColourSpellFromKey(key)
local function ColourSpellFromKey(key)
	local name = key.name or key.list[1]
	local c = key.colour or {1, 1, 1}

	local colour = format("%02X%02X%02X", c[1] * 255, c[2] * 255, c[3] * 255)

	if (wow24) then
		local icon = BS:GetSpellIcon(key.list[1])
		if (icon) then
			return format("|T%s:14|t|cFF%s%s|r", colour, name), colour
		end
	end

	return format("|cFF%s%s|r", colour, name), colour
end

-- ToggleKeyType
local function ToggleKeyType(keytype, onoff)
	if (onoff == nil) then
		onoff = not template[keytype]
	end

	if (onoff) then
		local dup = zg.buffs[keytype].dup
		if (dup) then
			for j,z in pairs(zg.buffs) do
				if (z.dup == dup) then
					zg:ModifyTemplate(j, nil)
				end
			end
		end

		zg:ModifyTemplate(keytype, true)
	else
		zg:ModifyTemplate(keytype, nil)
	end
	z:SetupForSpell()
	zg:CheckBuffs()
end

-- SetMenu
function zg:MakeSpellOptions()
	if (not self.options.args.spells) then
		self.options.args.spells = {
			type = "group",
			name = L["Group Spells"],
			desc = L["Group spell configuration"],
			order = 1,
			args = {
			}
		}
	end

	local args = self.options.args.spells.args

	local list = new()
	for k,v in pairs(self.buffs) do
		list[v.o] = v
	end

	for i = 1,#list do
		local key = list[i]
		local name = list[i].name or list[i].list[1]
		local cName = ColourSpellFromKey(key)
		args[name] = {
			type = "group",
			name = cName,
			desc = cName,
			order = i,
			isChecked = function(k) return template[key.type] end,
			onClick = ToggleKeyType,
			passValue = key.type,
			args = {
				header = {
					type = "header",
					name = cName,
					order = 1,
				},
				rebuff = {
					type = "range",
					name = L["Expiry Prelude"],
					desc = format(L["Rebuff prelude for %s (0=Module default)"], cName),
					func = timeFunc,
					order = 2,
					get = getPrelude,
					set = setPrelude,
					passValue = key.type,
					min = 0,
					max = 30 * 60,
					step = 15,
					bigStep = 60,
					order = 2,
				}
			}
		}
	end
end

-- SelectTemplate
function zg:OnSelectTemplate(templateName)
	template = self.db.char.templates.current
end

-- GetBuffedMembers
function zg:GetBuffedMembers()
	local temp = new()
	local totals = new(0, 0, 0, 0, 0, 0, 0, 0)
	local groupTotal = new(0, 0, 0, 0, 0, 0, 0, 0)
	local groupPresent = new(0, 0, 0, 0, 0, 0, 0, 0)
	local skipGroups = new()
	local dbGroups = self.db.char.groups
	local minTimeLeft
	local totalMembers, totalPresent = 0, 0
	local anyBlacklisted

	for unit in roster:IterateRoster() do
		local subgroup = z:GetGroupNumber(unit.unitid)
		local unitid = unit.unitid
		local grp = subgroup
		anyBlacklisted = anyBlacklisted or z:IsBlacklisted(unit.name)
		totalMembers = totalMembers + 1
		groupTotal[grp] = groupTotal[grp] + 1
		if (dbGroups[grp] and unit.online and UnitIsVisible(unitid) and not UnitIsDeadOrGhost(unitid) and not UnitIsCharmed(unit)) then
			local addSpecial
			if (CheckInteractDistance(unit.unitid, 1)) then
				totalPresent = totalPresent + 1
				groupPresent[grp] = groupPresent[grp] + 1
			else
				addSpecial = true
			end

			totals[grp] = totals[grp] + 1
			for i = 1,40 do
				local name, rank, buff, count, max, dur = UnitBuff(unitid, i)
				if (not name) then
					break
				end

				local buff = self.lookup[name]
				if (buff) then
					if (template[buff.type]) then
						local requiredTimeLeft = (self.db.char.rebuff and self.db.char.rebuff[buff.type]) or self.db.char.rebuff.default
						if ((not requiredTimeLeft or not dur or dur > requiredTimeLeft) and (buff.onlyManaUsers and (unit.class == "ROGUE" or unit.class == "WARRIOR"))) then
							if (addSpecial) then
								-- Pretend this person is in range, even though they're not cos they don't need this buff
								totalPresent = totalPresent + 1
								groupPresent[grp] = groupPresent[grp] + 1
							end
							if (not temp[buff.type]) then
								temp[buff.type] = new(0, 0, 0, 0, 0, 0, 0, 0)
							end
							temp[buff.type][grp] = temp[buff.type][grp] + 1
						--else
						--	ChatFrame2:AddMessage("---- Skipping buff with only "..tostring(dur).." left on it because it is less than the required "..tostring(requiredTimeLeft))
						end

						if (dur and (not minTimeLeft or dur - requiredTimeLeft < minTimeLeft)) then
							minTimeLeft = dur - requiredTimeLeft
						end
					end
				end
			end
		end
	end

	z.waitingForClass = nil
	if (z.db.profile.waitforclass) then
		for i = 1,8 do
			if (groupTotal[i] > 0 and groupPresent[i] < groupTotal[i]) then
				z.waitingForClass = (z.waitingForClass or "") .. ((z.waitingForClass and ", ") or "") .. format(L["Group %d"], i)
				skipGroups[i] = true
			end
		end
	end

	del(groupPresent)
	del(groupTotal)

	return temp, totals, minTimeLeft, totalPresent / totalMembers, skipGroups, anyBlacklisted
end

-- FindUnitInRangeMissing
function zg:FindUnitInRangeMissing(group, typ)
	if (not self.buffs[typ]) then
		error(format("Unrecognised type %q", tostring(typ)))
	end
	local list = self.buffs[typ].list
	local rangeCheckSpell = list[1]

	local requiredTimeLeft = (self.db.char.rebuff and self.db.char.rebuff[typ]) or self.db.char.rebuff.default

	for unit in roster:IterateRoster() do
		local subgroup = z:GetGroupNumber(unit.unitid)

		if (subgroup == group and not z:IsBlacklisted(unit.name)) then
			local unitid = unit.unitid
			if (unit.online and UnitIsVisible(unitid) and not UnitIsDeadOrGhost(unitid) and not UnitIsCharmed(unitid) and IsSpellInRange(rangeCheckSpell, unitid) == 1 and ((not z.db.profile.skippvp or not UnitIsPVP(unitid)) or UnitIsPVP("player"))) then
				local got
				for i = 1,40 do
					local name, rank, buff, count, max, dur = UnitBuff(unitid, i)
					if (not name) then
						break
					end

					local buff = self.lookup[name]
					if (buff and buff.type == typ) then
						if (dur and dur < requiredTimeLeft) then
							return unitid
						end
						got = true
						break
					end
				end

				if (not got) then
					local t = self.buffs[typ]
					if (not t.onlyManaUsers or (unit.class ~= "ROGUE" and unit.class ~= "WARRIOR")) then
						return unitid
					end
				end
			end
		end
	end
end

-- CheckBuffs
function zg:CheckBuffs()
	if (not self.db or not z:CanCheckBuffs()) then
		return
	end

	-- First build a list of total group counts, and total numbers of players who have each buff in each group
	local temp, totals, minTimeLeft, percentPresent, skipGroups, anyBlacklisted = self:GetBuffedMembers()
	local skipClassBuffs = z.db.char.singlesInBG and select(2, IsInInstance()) == "pvp"

	if (z.db.profile.waitforraid > 0 and not skipClassBuffs) then
		-- See if enough of raid present
		if (percentPresent < z.db.profile.waitforraid) then	-- Wait for % of raid before buffing
			z.waitingForRaid = floor(percentPresent * 100)
			self:ScheduleEvent("ZOMGBuffTehRaid_CheckBuffs", self.CheckBuffs, 5, self)
			return
		end
	end
	z.waitingForRaid = nil

	-- Now go through this and find anything we need to buff
	local dbGroups = self.db.char.groups
	local any
	for i = 1,8 do								-- i == group
		if (dbGroups[i] and not skipGroups[i]) then
			local players = totals[i]
			if (players > 0) then
				for k,v in pairs(template) do			-- k = spellType (INT, STA etc.), v = buff data for that type
					if (k ~= "modified" and k ~= "state") then
						local typeSpec = self.buffs[k]
						local gotBuff = (temp[k] and temp[k][i]) or 0
						if (not gotBuff or gotBuff < players) then
							local missingBuff = players - gotBuff

							local unit = self:FindUnitInRangeMissing(i, k)
							if (unit) then
								local buffSpec = typeSpec.list
								local t1, t2
								if (buffSpec[2]) then
									t1, t2 = IsUsableSpell(buffSpec[2])
								end
								if (skipClassBuffs or (missingBuff < self.db.char.groupcast or not (t1 or t2))) then
									-- Do single buff
									z:Notice(L["%s needs %s"], z:ColourUnit(unit), typeSpec.name or buffSpec[1])
									z:SetupForSpell(unit, buffSpec[1], self)
									any = true
									break
								else
									-- Do group buff
									z:Notice(L["Group %d needs %s"], i, typeSpec.name or buffSpec[1])
									z:SetupForSpell(unit, buffSpec[2], self)
									any = true
									break
								end
							--else
								--error("Know we're supposed to buff somoene, but can't find anyone atm")
							end
						end
					end
				end
				if (any) then
					break
				end
			end
		end
	end

	if (not any and z.db.profile.buffpets) then
		-- Catch any pets that missed the group buffing
		for unit in roster:IterateRoster(true) do
			local unitid = unit.unitid
--if (unit.class == "PET") then
--self:Print("Pet found: unit.subgroup = "..tostring(unit.subgroup))
--end
			local ownerunitid
			if (unitid == "pet") then
				ownerunitid = "player"
			else
				ownerunitid = unitid:gsub("pet", "")
			end
			local ownerunit = roster:GetUnitObjectFromUnit(ownerunitid)
			if (ownerunit and unit.class == "PET" and UnitIsVisible(unitid) and UnitCanAssist("player", unitid) and CheckInteractDistance(unitid, 1)) then
				local subgroup = z:GetGroupNumber(ownerunit)

 				if (self.db.char.groups[subgroup]) then
					for i = 1,40 do
						local name, rank, buff, count, max, dur = UnitBuff(unitid, i)
						if (not name) then
							break
						end

						local buff = self.lookup[name]
						if (buff) then
							if (template[buff.type]) then
								local requiredTimeLeft = (self.db.char.rebuff and self.db.char.rebuff[buff.type]) or self.db.char.rebuff.default
								if ((not requiredTimeLeft or not dur or dur > requiredTimeLeft) and (buff.onlyManaUsers and (unit.class == "ROGUE" or unit.class == "WARRIOR"))) then
									z:Notice(L["%s needs %s"], z:ColourUnit(unitid), buff.name)
									z:SetupForSpell(unit, buff.list[1], self)
									any = true
									break
								end

								if (dur and (not minTimeLeft or dur - requiredTimeLeft < minTimeLeft)) then
									minTimeLeft = dur - requiredTimeLeft
								end
							end
						end
					end
					if (any) then
						break
					end
				end
			end
		end
	end

	self:CancelScheduledEvent("ZOMGBuffTehRaid_CheckBuffs")
	if (anyBlacklisted) then
		minTimeLeft = 1.5
	end
	if (not any) then
		if (z.waitingForClass) then
			self:ScheduleEvent("ZOMGBuffTehRaid_CheckBuffs", self.CheckBuffs, 5, self)
		else
			self:ScheduleEvent("ZOMGBuffTehRaid_CheckBuffs", self.CheckBuffs, minTimeLeft or 60, self)
		end
	end

	deepDel(temp)
	deepDel(groups)
	del(totals)
	del(skipGroups)
end

-- UNIT_AURA
function zg:UNIT_AURA(unit)
	if (UnitInParty(unit) or UnitInRaid(unit)) then
		-- If we have a buff loaded up, then we need to see if another of our class has buffed this person
		z:CheckForChange(self)
	end
end

-- GetActions
function zg:GetActions()
	if (not self.actions) then
		self.actions = {}

		local list = new()
		for k,v in pairs(self.buffs) do			-- Order them into 1, 2, 3
			list[v.o] = v
		end

		local t1, t2
		for i,v in ipairs(list) do
			t1, t2 = IsSpellInRange(v.list[1], "player")	-- Rather than IsUsableSpell will assign when we don't have reagents but do have spell
			if (t1 or t2) then
				if (v.keycode) then
					if (v.list[2]) then
						for j = 1,#v.list do
							tinsert(self.actions, {name = v.list[j], type = v.keycode..j})
						end
					else
						tinsert(self.actions, {name = v.list[1], type = v.keycode})
					end
				end
			end
		end

		del(list)
	end
	return self.actions
end

-- ResetActions
function zg:ResetActions()
	self.actions = nil
end

-- RebuffQuery
function zg:RebuffQuery(unit)
	if (self.db and template and UnitIsConnected(unit) and UnitIsVisible(unit) and not UnitIsDeadOrGhost(unit) and not UnitIsCharmed(unit)) then
		for i = 1,40 do
			local name, rank, buff, count, max, dur = UnitBuff(unit, i)
			if (not name) then
				break
			end

			if (name == mainSingle or name == mainGroup) then
				return
			end
		end

		for buffType,stuff in pairs(self.buffs) do
			if (stuff.required and template[buffType]) then
				if (stuff.onlyManaUsers and (UnitClass(unit) == "ROGUE" or UnitClass(unit) == "WARRIOR")) then
					return
				end
			end
		end

		return true
	end
end

-- OnModuleInitialize
function zg:OnModuleInitialize()
	playerClass = select(2, UnitClass("player"))

	if (playerClass == "PRIEST") then
		local t1, t2 = IsUsableSpell(BS["Divine Spirit"])
		self.buffs = {
			STA = {
				o = 1,
				name = SPELL_STAT3_NAME,
				list = {BS["Power Word: Fortitude"], BS["Prayer of Fortitude"]},
				reagents = {nil, {
						[1] = R["Holy Candle"],
						[2] = R["Sacred Candle"],
						[3] = R["Sacred Candle"],
					},
				},
				required = true,
				colour = {0.7, 0.7, 1},
				keycode = "stamina",
			},
			SPIRIT = {
				o = 2,
				name = SPELL_STAT5_NAME,
				list = {BS["Divine Spirit"], BS["Prayer of Spirit"]},
				reagents = {nil, R["Sacred Candle"]},
				required = t1 or t2,
				onlyManaUsers = true,
				colour = {0.3, 0.5, 1},
				keycode = "spirit",
			},
			SHADOWPROT = {
				o = 3,
				list = {BS["Shadow Protection"], BS["Prayer of Shadow Protection"]},
				reagents = {nil, R["Sacred Candle"]},
				required = false,
				colour = {0.7, 0.2, 0.7},
				keycode = "shadowprot",
			}
		}
		self.reagents = {
			[R["Holy Candle"]] = {0, 1, 500, maxLevel = 59},
			[R["Sacred Candle"]] = {0, 1, 500},
		}

	elseif (playerClass == "DRUID") then
		self.buffs = {
			MARK = {
				o = 1,
				name = L["Mark"],
				list = {BS["Mark of the Wild"], BS["Gift of the Wild"]},
				reagents = {nil, {
						[1] = R["Wild Berries"],
						[2] = R["Wild Thornroot"],
						[3] = R["Wild Quillvine"],
					},
				},
				required = true,
				colour = {0.9, 0.5, 0.9},
				keycode = "mark",
			},
			THORNS = {
				o = 2,
				list = {BS["Thorns"]},
				required = false,
				colour = {0.7, 0.7, 0.3},
				keycode = "thorns",
			}
		}
		self.reagents = {
			[R["Wild Berries"]] = {0, 1, 500},
			[R["Wild Thornroot"]] = {0, 1, 500},
			[R["Wild Quillvine"]] = {0, 1, 500}
		}

	elseif (playerClass == "MAGE") then
		self.buffs = {
			INT = {
				o = 1,
				name = SPELL_STAT4_NAME,
				list = {BS["Arcane Intellect"], BS["Arcane Brilliance"]},
				reagents = {nil, R["Arcane Powder"]},
				required = true,
				onlyManaUsers = true,
				colour = {0.3, 0.3, 1},
				keycode = "int"
			},
			DAMPEN = {
				o = 2,
				list = {BS["Dampen Magic"]},
				required = false,
				dup = 1,
				colour = {0.3, 1, 0.8},
				keycode = "dampen",
			},
			AMPLIFY = {
				o = 3,
				list = {BS["Amplify Magic"]},
				required = false,
				dup = 1,
				colour = {1, 1, 0.25},
				keycode = "amplify",
			}
		}
		self.reagents = {
			[R["Arcane Powder"]] = {(UnitLevel("player") >= 56 and 20) or 0, 1, 500},
		}
	end

	if (not self.buffs) then
		return
	end

	z.SetClickSpells = function(self, cell)
		if (zg.actions) then
			for i,action in ipairs(zg.actions) do
				local Mod, button = z:GetActionClick(action.type)
				if (button) then
					z:SetACellSpell(cell, Mod, button, action.name)
				end
			end
		end
	end

	local defaultForClass = {Default = {}}
	for k,v in pairs(self.buffs) do
		if (v.required) then
			defaultForClass.Default[k] = true
		end
	end

	self.db = z:AcquireDBNamespace("BuffTehRaid")
	z:RegisterDefaults("BuffTehRaid", "char", {
		reagents = {},
		templates = defaultForClass,
		defaultTemplate = "Default",
		groupcast = 2,
		groups = {true, true, true, true, true, true, true, true},
		autogroups = true,
		rebuff = {
			default = 30,
		},
	} )
	z:RegisterChatCommand({"/zomgraid", "/zomgbufftehraid"}, zg.options)
	self.OnMenuRequest = self.options
	z.options.args.ZOMGBuffTehRaid = self.options

	self.lookup = {}
	self.required = 0
	for k,v in pairs(self.buffs) do
		v.type = k
		for i,buff in pairs(v.list) do
			self.lookup[buff] = v
		end
	end

	z.OnCommReceive.AUTOGROUPASSIGNED = function(self, prefix, sender, channel, class, displayList)
		if (not zg.db.char.autogroups and class == playerClass) then
			if (not zg.autoGroupWarned) then
				zg.autoGroupWarned = true
				self:Print(L["Warning: %s has auto-assigned themselves to buff groups %s, but you have the Auto Group Assignment option disabled"], z:ColourUnitByName(sender), table.concat(displayList, ", "))
			end
		end
	end

	self.OnModuleInitialize = nil
end

-- AutoGroupAssignment
function zg:AutoGroupAssignment()
	if (not self.db or not self.db.char.autogroups) then
		return
	end

	local count = z.classcount[playerClass]
	local list = new()
	local maxGroup = 1
	if (GetNumRaidMembers() > 0) then
		for i = 1,GetNumRaidMembers() do
			local name, rank, subgroup, level, _, class = GetRaidRosterInfo(i)
			if (class == playerClass) then
				tinsert(list, name)
			end
			if (subgroup > maxGroup) then
				maxGroup = subgroup
			end
		end
	end
	sort(list)

	local myIndex = 1
	for i,name in ipairs(list) do
		if (name == playerName) then
			myIndex = i
			break
		end
	end

	local displayList = new()
	if (maxGroup == 1 or count == 1) then
		for i = 1,8 do
			self.db.char.groups[i] = true
		end
		z:CheckForChange(self)
		return
	end

	local anyChange
	local each = maxGroup / count
	local start = (myIndex - 1) * each + 1
	for groupNo = 1,maxGroup do
		local shouldBuff = groupNo >= start and groupNo < start + each
		if (shouldBuff) then
			tinsert(displayList, tostring(groupNo))
		end
		if (shouldBuff ~= self.db.char.groups[groupNo]) then
			self.db.char.groups[groupNo] = shouldBuff
			anyChange = true
		end
	end
	if (not next(displayList)) then
		for i = 1,8 do
			self.db.char.groups[i] = true
		end
		z:CheckForChange(self)
		return
	end

	if (anyChange) then
		if (GetNumRaidMembers() > 0 and z.db.profile.info) then
			self:Print(L["You are now responsible for Groups %s"], table.concat(displayList, ", ", 1, max(1, #displayList - 1))..((#displayList > 1 and L[" and "]..displayList[#displayList]) or ""))
		end
		z:CheckForChange(self)
		z:SendCommMessage("GROUP", "AUTOGROUPASSIGNED", playerClass, displayList)
	end

	del(list)
	del(displayList)
end

-- CheckRoster
function zg:CheckRoster()
	if (self.db) then
		self:AutoGroupAssignment()
	end
end

-- ShowBuffBar
function zg:ShowBuffBar(cell, name)
	local prefix = SecureButton_GetModifierPrefix(cell)
	for button = 1,5 do
		local spell = cell:GetAttribute(prefix, "spell", button)
		if (spell == name) then
			return true
		else
			local find = self.lookup[spell]
			if (find and find.list[2]) then
				if (find.list[1] == name or find.list[2] == name) then
					return true
				end
			end
		end
	end
end

-- GetSpellColour
function zg:GetSpellColour(spellName)
	local def = self.lookup and self.lookup[spellName]
	if (def) then
		local c = def.colour
		return format("%02X%02X%02X", c[1] * 255, c[2] * 255, c[3] * 255)
	end
end

-- OneOfYours
function zg:OneOfYours(spell)
	return self.lookup[spell]
end

-- SayWhatWeDid
function zg:SayWhatWeDid(icon, spell, name, rank)
	local s = spell or icon:GetAttribute("spell")
	if (s) then
		local found = self.lookup[s]
		if (found) then
			local _, colour = ColourSpellFromKey(found)
			local unit = roster:GetUnitObjectFromName(name)
			if (unit and found.list and spell == found.list[2]) then
				local r = found.reagents and found.reagents[2]
				local subgroup = z:GetGroupNumber(unit.unitid)
				if (r) then
					local reagent
					if (type(r) == "table") then
						if (r[rank]) then
							reagent = r[rank]
						else
							regent = r[#r]
						end
					else
						reagent = r
					end

					local count = GetItemCount(reagent)
					local colourCount
					if (count < 5) then
						colourCount = "|cFFFF4040"
					elseif (count < 10) then
						colourCount = "|cFFFFFF40"
					else
						colourCount = "|cFF40FF40"
					end

					self:Print(L["%s on Group %d (%s%d|r)"], z:LinkSpell(s, colour), subgroup, colourCount, count - 1)
				else
		      		self:Print(L["%s on Group %d"], z:LinkSpell(s, colour), subgroup)
				end
			else
				self:Print(L["%s on %s"], z:LinkSpell(s, colour), z:ColourUnitByName(name))
			end
		end
	end
end

-- UNIT_SPELLCAST_FAILED
function zg:SpellCastFailed(spell, name, manual)
	if (not manual) then
		if (self.lookup[spell]) then
			z:Blacklist(name)
		end
	end
end

-- SPELLS_CHANGED
function zg:SPELLS_CHANGED()
	self.actions = nil
	z:UpdateCellSpells()
	--self:GetMainSpells()

	for name,spellDef in pairs(self.buffs) do
		if (spellDef.o == 1) then
			mainSingle = spellDef.list[1]
			mainGroup = spellDef.list[2]
			break
		end
	end

	z:CheckForChange(self)
end

-- TooltipUpdate
function zg:TooltipUpdate(cat)
	if (template) then
		cat:AddLine('text', " ")
		cat:AddLine(
			"text", L["Group Template: "].."|cFFFFFFFF"..(self.db.char.selectedTemplate or L["none"]),
			"text2", (template and template.modified and "|cFFFF4040"..L["(modified)"].."|r") or ""
		)

		local any
		for k,v in pairs(self.buffs) do
			local key = self.buffs[k]
			local endis
			if (template[key.type]) then
				endis = "|cFF80FF80"..L["Enabled"].."|r"
			else
				endis = "|cFFFF8080"..L["Disabled"].."|r"
			end

			checkIcon = BS:GetSpellIcon(key.list[1])

			cat:AddLine(
				"text", ColourSpellFromKey(key),
				"text2", endis,
				"func", ToggleKeyType,
				"arg1", key.type,
				"hasCheck", checkIcon and true,
				"checked", checkIcon and true,
				"checkIcon", checkIcon
			)
		end
	end
end

-- OnResetDB
function zg:OnResetDB()
	if (self.db) then
		template = self.db.char.templates.current

		for k,v in pairs(template) do
			if (k ~= "modified" and k ~= "state") then
				if (not self.buffs[k]) then
					template[k] = nil
				end
			end
		end
	end
end

-- OnModuleEnable
function zg:OnModuleEnable()
	if (self.db) then
		local class = select(2, UnitClass("player"))
		if (class ~= playerClass and self.OnModuleInitialize) then
			self:OnModuleInitialize()
		end
	
		z:MakeOptionsReagentList()
		self:OnResetDB()
		
		self:SPELLS_CHANGED()
		self:MakeSpellOptions()

		self:RegisterBucketEvent("UNIT_AURA", 0.2)				-- We don't care who
		self:RegisterBucketEvent("SPELLS_CHANGED", 2)
		z:CheckForChange(self)
	end
end

-- OnDisable
function zg:OnDisable()
	z:CheckForChange(self)
end
